/* $FILENAME$
 * $AUTHORS$, Tue Jun 6 2006
 *
 */


#define USEGRID 0


#include <GL/glut.h>
#include <cstdio>
#include <vector>
#include "image.h"
#include "imageio.h"
#include "displayimage.h"
#if USEGRID
#include "grid.h"
#include "displaygrid.h"
#include "griduiaction.h"
#include "deformimagewithgrid.h"
#else
#include "deformimagewithcontrolpoint.h"
#include "controlpoint.h"
#include "controlpointuiaction.h"
#include "displaycontrolpoint.h"
#endif

Image image; // the image
ImageIO<Image> io(&image); // load an image
//DisplayImage<Image> displayImage(&image); // display the original image

#if USEGRID
Grid grid(4, 4); // the grid with MxN control points
DisplayGrid displayGrid(&grid); // display the grid
GridUIAction gridUI(&grid); // deform the grid
DeformImageWithGrid deformedImage(&image, &grid); // deform image with grid
DisplayImage<DeformImageWithGrid> displayImage(&deformedImage); // display the deformed image
#else
std::vector<ControlPoint> controlPoints; // control points
DisplayControlPoint displayControlPoints(&controlPoints); // display the control points
ControlPointUIAction controlPointsUI(&controlPoints); // move control points
DeformImageWithControlPoint deformedImage(&image, &controlPoints); // deform image with control points
DisplayImage<DeformImageWithControlPoint> displayImage(&deformedImage); // display the deformed image
#endif

int imageSize[2] = { 0,0 }; // image size
float winratio[2] = { 1,1 }; // imagesize / window size, used when resizing the window
int winid = 0; // window id, so we can destroy it later

void init()
{
	// Initialize display
	glClearColor(0,0,0,0);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0, 0, imageSize[0], imageSize[1]);
	glOrtho(0,imageSize[0],
		0,imageSize[1],
		-1,+1);

#if USEGRID
	// Initialize grid
	float spacing[2] = { (float)imageSize[0]/(grid.getSize()[0]-1), (float)imageSize[1]/(grid.getSize()[1]-1) };
	grid.initOrthogonal(spacing);
#endif

	// Deform image
	deformedImage.update();
}

void display()
{
	// Erase buffer
	glClear(GL_COLOR_BUFFER_BIT);
	glClearColor(0,0,0,1);

	// Display the deformed image
	deformedImage.update(); // deform image at each redisplay
	displayImage.display();
	
#if USEGRID
	// Display grid
	displayGrid.display();
#else
	// Display control points
	displayControlPoints.display();
#endif

	glutSwapBuffers();
}


void reshape(int w, int h)
{
	glViewport(0, 0, w, h);
	winratio[0] = (float)imageSize[0]/w;
	winratio[1] = (float)imageSize[1]/h;

	glutSwapBuffers();
}


void mouse(int button, int state, int x, int y)
{
	switch(button)
	{
		// If a button is pressed, start the motion of a control point
		case GLUT_LEFT_BUTTON:
		case GLUT_RIGHT_BUTTON:
		case GLUT_MIDDLE_BUTTON:
			switch(state)
			{
			case GLUT_DOWN:
			{
				float pos[2] = { x*winratio[0], (imageSize[1]/winratio[1]-y)*winratio[1] };
#if USEGRID
				// Start the motion of a control point
				gridUI.beginMotion(pos);
#else
				// Select a control point, otherwise add a new point
				if(!controlPointsUI.selectControlPoint(pos[0], pos[1], 10))
					controlPointsUI.addControlPoint(pos[0], pos[1]);				
#endif
				break;
			}
			case GLUT_UP:
			{
				deformedImage.update();
				glutPostRedisplay();
				break;
			}
			default: break;
			};
			break;
			
		default: break;
	}
}


void motion(int x, int y)
{
	float pos[2] = { x*winratio[0], (imageSize[1]/winratio[1]-y)*winratio[1] };
	
#if USEGRID
	// Deform grid
	gridUI.motion(pos);

	// Display grid
	displayGrid.display();
#else
	// Move selected control point
	controlPointsUI.motion(pos);
#endif
	
	glutPostRedisplay();	
}


void keyboard(unsigned char key, int x, int y)
{
	switch(key)
	{
		case 'w':
		{
#if USEGRID
			ImageIO<DeformImageWithGrid> io(&deformedImage);
#else
			ImageIO<DeformImageWithControlPoint> io(&deformedImage);
#endif
			io.writePGM("result.pgm");
			break;
		}
		case 'q':
		{
			glutDestroyWindow(winid); // close window, terminate glut loop
			break;
		}
	};
}


int main(int argc, char *argv[])
{
	// Check arguments
	if(argc!=2) {
		printf("program filename.pgm\n");
		return 1;
	}
	
	// Load image
	io.readPGM(argv[1]);
	imageSize[0] = image.getSize()[0];
	imageSize[1] = image.getSize()[1];
	
	// Initialize window
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(imageSize[0], imageSize[1]);
	glutInitWindowPosition(100,100);
	winid = glutCreateWindow("Thin Plate");
	init();
	
	// Ininitialize callbacks
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	
	// Main loop
	glutMainLoop();
	
	return 0;
 }
