/* thinplateinterpolator
 * $AUTHORS$, Tue Jun 6 2006
 *
 * Reusing code from the Numerical Recipes
 *
 * Reusing code from http://elonen.iki.fi/code/tpsdemo/index.html
 * Copyright (C) 2003, 2004, 2005 by Jarno Elonen
 */
#include "thinplatespline.h"
#include <cmath>


// Numerical Recipe to solver linear equation with LU decomposition
void ludcmp(double **a, int n, int *indx, double *d);
void lubksb(double **a, int n, int *indx, double b[]);
	
ThinPlateSpline::ThinPlateSpline(int nbCellsX, int nbCellsY)
{
	// Number of cells in x and y direction
	// Cells are used as discretization of the surface between 0,0 and 1,1
	m_nbCells[0] = nbCellsX;
	m_nbCells[1] = nbCellsY;
	
	// Allocate memory for the cells
	float* contiguous = new float [nbCellsX*nbCellsY];
	m_grid = new float* [nbCellsY];
	for(int j=0; j<nbCellsY; ++j)
		m_grid[j] = &contiguous[j*nbCellsX];
	
	// Initialize them at 0 (no displacement)
	for(int j=0; j<nbCellsY; ++j)
		for(int i=0; i<nbCellsX; ++i)
			m_grid[j][i] = 0;
}


ThinPlateSpline::~ThinPlateSpline()
{
	// Delete cells
	delete [] *m_grid; // contiguous memory
	delete [] m_grid; // row pointers
}


// Set the control points and interpolate the grid displacement, return the bending energy
double ThinPlateSpline::setControlPoints(std::vector<ControlPoint> controlPoints)
{
	double regularization = 0.0;
	double bending_energy = 0.0;

/*
	*  Calculate Thin Plate Spline (TPS) weights from
	*  control points and build a new height grid by
	*  interpolating with them.
 */
	// You We need at least 3 points to define a plane
	if ( controlPoints.size() < 3 )
		return -1;

	int p = controlPoints.size();

	// Allocate the matrix and vector
	double* c1 = new double [(p+3)*(p+3)]; // contiguous memory
	double* c3 = new double [(p)*(p)]; // contiguous memory
	double** mtx_l = new double* [p*3]; for(int i=0; i<p*3; ++i) mtx_l[i] = &c1[i*(p+3)];
	double*  mtx_v = new double  [p*3];
	double** mtx_orig_k = new double* [p]; for(int i=0; i<p; ++i) mtx_orig_k[i] = &c3[i*(p)];

	// Fill K (p x p, upper left of L) and calculate
	// mean edge length from control points
	//
	// K is symmetrical so we really have to
	// calculate only about half of the coefficients.
	double a = 0.0;
	for ( int i=0; i<p; ++i )
	{
		for ( int j=i+1; j<p; ++j )
		{
			ControlPoint pt_i = controlPoints[i];
			ControlPoint pt_j = controlPoints[j];
			pt_i.h = pt_j.h = 0;
			double elen = sqrt((pt_i.x-pt_j.x)*(pt_i.x-pt_j.x) + // length of pt_i-pt_j
						(pt_i.y-pt_j.y)*(pt_i.y-pt_j.y));
			mtx_l[i][j] = mtx_l[j][i] =
				mtx_orig_k[i][j] = mtx_orig_k[j][i] =
				tps_base_func(elen);
			a += elen * 2; // same for upper & lower tri
		}
	}
	a /= (double)(p*p);

	// Fill the rest of L
	for ( int i=0; i<p; ++i )
	{
		// diagonal: reqularization parameters (lambda * a^2)
		mtx_l[i][i] = mtx_orig_k[i][i] =
			regularization * (a*a);

		// P (p x 3, upper right)
		mtx_l[i][p+0] = 1.0;
		mtx_l[i][p+1] = controlPoints[i].x;
		mtx_l[i][p+2] = controlPoints[i].y;

		// P transposed (3 x p, bottom left)
		mtx_l[p+0][i] = 1.0;
		mtx_l[p+1][i] = controlPoints[i].x;
		mtx_l[p+2][i] = controlPoints[i].y;
	}
	// O (3 x 3, lower right)
	for ( int i=p; i<p+3; ++i )
		for ( int j=p; j<p+3; ++j )
			mtx_l[i][j] = 0.0;


	// Fill the right hand vector V
	for ( int i=0; i<p; ++i )
		mtx_v[i] = controlPoints[i].h;
	mtx_v[p+0] = mtx_v[p+1] = mtx_v[p+2] = 0.0;

	// Solve the linear system "inplace"
	int* indx = new int [p+3]; double d;
	ludcmp(mtx_l, p+3, indx, &d); // LU decomposition
	lubksb(mtx_l, p+3, indx, mtx_v); // LU Solving
	delete [] indx;

	// Interpolate grid heights
	float spacing[2] = { 1.0f/(m_nbCells[0]-1), 1.0f/(m_nbCells[1]-1) };
	for ( int y=0; y<m_nbCells[1]; ++y )
		for ( int x=0; x<m_nbCells[0]; ++x )
	{
		{
			double h = mtx_v[p+0] + mtx_v[p+1]*x*spacing[0] + mtx_v[p+2]*y*spacing[1];
			ControlPoint pt_i, pt_cur(x*spacing[0],y*spacing[1],0);
			for ( int i=0; i<p; ++i )
			{
				pt_i = controlPoints[i];
				pt_i.h = 0;
				double elen = sqrt((pt_i.x-pt_cur.x)*(pt_i.x-pt_cur.x) + // length of pt_i-pt_cur
							(pt_i.y-pt_cur.y)*(pt_i.y-pt_cur.y));
				h += mtx_v[i] * tps_base_func(elen);
			}
			m_grid[y][x] = h;
		}
	}

	// Calc bending energy
	//matrix<double> w( p, 1 );
	//for ( int i=0; i<p; ++i )
	//	w(i,0) = mtx_v(i,0);
	//matrix<double> be = prod( prod<matrix<double> >( trans(w), mtx_orig_k ), w );
	//bending_energy = be(0,0);
	
	// Calc bending energy (doing the product right here)
	double *w = new double [p];
	double **prod = mtx_l; // reuse mtx_l memory
	for ( int i=0; i<p; ++i )
		w[i] = mtx_v[i];
	// product transw*mtx_orig_k
	for(int j=0; j<p; ++j)
		for(int i=0; i<p; ++i)
	{
		prod[j][i] = 0;
		for(int k=0; k<p; ++k)
			prod[j][i] += w[k] * mtx_orig_k[k][i];
	}
	// product prod w
	double** prod2 = mtx_orig_k; // reuse mtx_orig_k memory
	for(int j=0; j<p; ++j)
		for(int i=0; i<p; ++i)
	{
		prod2[j][i] = 0;
		for(int k=0; k<p; ++k)
			prod2[j][i] += prod[j][k] * w[k];
	}
	bending_energy = prod2[0][0];

	// Delete from pointers
	delete [] *mtx_l; // contiguous memory
	delete [] mtx_l;
	delete [] mtx_v; // contiguous memory
	delete [] *mtx_orig_k; // contiguous memory
	delete [] mtx_orig_k;
	
	return bending_energy;
}


// Get the surface height at grid position i,j
float ThinPlateSpline::get(int i, int j) const
{
	if(i<0 || i>=m_nbCells[0] || j<0 || j>=m_nbCells[1]) {
		puts( "index out of reach" );
		exit(1);
	}
	return m_grid[j][i];
}


// Get the surface height at position x,y (x,y between 0 and 1)
// If (x,y) is between two grid node, the height is interpolated
// using a bilinear interpolation
float ThinPlateSpline::get(float x, float y) const
{
	// Bilinear interpolation
	// Q01 --R1---- Q11
	//  |    |       |
	//  |----P-------|
	//  |    |       |
	// Q00 --R0---- Q10
	// f(P) = (y1-y)/(y1-y0) f(R0) + (y-y0)/(y1-y0) f(R1)
	//  with f(R0) = (x1-x)/(x1-x0) f(Q00) + (x-x0)/(x1-x0) f(Q10)
	//  with f(R1) = (x1-x)/(x1-x0) f(Q01) + (x-x0)/(x1-x0) f(Q11)
	// f(Q) is the intensity in the center of cell Q
	
	// Spacing of a cell
	float dx = 1.0f/(m_nbCells[0]-1);
	float dy = 1.0f/(m_nbCells[1]-1);
	
	// Find grid index of Q00
	int i = (int)(x/dx);
	int j = (int)(y/dy);
	if(i>=m_nbCells[0]) i = m_nbCells[0]-1;
	if(j>=m_nbCells[1]) j = m_nbCells[1]-1;
	
	// First dimension
	float r0 = ((i+1)*dx-x)/dx*m_grid[j][i] + (x-i*dx)/dx*m_grid[j][i+1];
	float r1 = ((i+1)*dx-x)/dx*m_grid[j+1][i] + (x-i*dx)/dx*m_grid[j+1][i+1];
	
	// Second dimension
	return ((j+1)*dy-y)/dy*r0 + (y-j*dy)/dy*r1; // p
}


double ThinPlateSpline::tps_base_func(double r)
{
	if ( r == 0.0 )
		return 0.0;
	else
		return r*r * log(r);
}

// LU decomposition from Numerical Recipe
void ludcmp(double **a, int n, int *indx, double *d)
{
	int i,imax=0,j,k;
	double big,dum,sum,temp;
	double *vv;
	vv=new double[n+1];
	*d=1.0;
	for (i=0;i<n;i++) {
		big=0.0;
		for (j=0;j<n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) { printf("Singular matrix in routine ludcmp"); return; }
		vv[i]=1.0/big;
	}
	for (j=0;j<n;j++) {
		for (i=0;i<j;i++) {
			sum=a[i][j];
			for (k=0;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<n;i++) {
			sum=a[i][j];
			for (k=0;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=0;k<n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=1e-20;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<n;i++) a[i][j] *= dum;
		}
	}
	delete [] vv;
}

// LU solver from Numerical Recipe
void lubksb(double **a, int n, int *indx, double b[])
{
	int i,ii=-1,ip,j;
	double sum;
	for (i=0;i<n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii>=0)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n-1;i>=0;i--) {
		sum=b[i];
		for (j=i+1;j<n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}
