/* thinplatespline
 * $AUTHORS$, Tue Jun 6 2006
 *
 * With some surface heights defined, interpolates any surface height of
 * a point (surface is defined between (0,0) and (1,1))
 * If the surface discretization is too small, surface heights of neighboring
 * grid nodes are used with a bilinear interpolation.
 * Uses a thin plate spline (see Bookstein).
 * Example of code:
 *
 *	typedef ThinPlateSpline::ControlPoint Point;
 *	std::vector<Point> controlPoints;
 *	controlPoints.push_back(Point(0,0,0)); // no height at corners
 *	controlPoints.push_back(Point(0,1,0));
 *	controlPoints.push_back(Point(1,0,0));
 *	controlPoints.push_back(Point(1,1,0));
 *	controlPoints.push_back(Point(0.5,0.5,0.5)); // height of 0.5 in the middle
 *	ThinPlateSpline surface(10,10); // surface has 10x10 cells
 *	surface.setControlPoints(controlPoints);
 *	float h = surface.get(0.25,0.25); // surface height at point(0.25,0.25)
 *
 * To interpolate the displacement of a point in 3 dimensions (dx,dy,dz)
 * use 3 interpolators (surfaces), one for each dimension
 *
 * Reusing code from the Numerical Recipes
 *
 * Reusing code from http://elonen.iki.fi/code/tpsdemo/index.html
 * Copyright (C) 2003, 2004, 2005 by Jarno Elonen
 */
 
#ifndef THINPLATESPLINE_H
#define THINPLATESPLINE_H

#include <vector>

#define DEFAULT_NB_CELLS 100
/**
	@author Herve Lombaert <herve.lombaert@polymtl.ca>
*/
class ThinPlateSpline{
public:
	ThinPlateSpline(int nbCellsX=DEFAULT_NB_CELLS, int nbCellsY=DEFAULT_NB_CELLS);
	~ThinPlateSpline();

	// A control point is defined by its current position and its displacement
	struct ControlPoint {
		float x,y; // positionn
		float h; // surface height
		ControlPoint(float x=0, float y=0, float h=0) : x(x),y(y),h(h) {}
	};
    
	// Set the control points and interpolate the grid displacement, return bending energy
	double setControlPoints(std::vector<ControlPoint> controlPoints);
    
	// Get the surface height at grid position i,j
	float get(int i, int j) const;
    
	// Get the surface height at position x,y (x,y between 0 and 1)
	// If (x,y) is between two grid node, the height is interpolated
	// using a bilinear interpolation
	float get(float x, float y) const;
    
private:
	float** m_grid; // surface discretization
	int m_nbCells[2]; // grid size
	
private:
	static double tps_base_func(double r);
};

#endif
