/* Timer.h
 * Herve Lombaert, July 9th, 2004
 * Timing tool
 */

#ifndef TIMER_H
#define TIMER_H

//#pragma warning (disable: 4786)

#include <vector>
#include <string>

class Timer
{
public:
	static Timer* Get();          // get instance of timer
	static void Destroy();        // destroy instance of timer

	void Stamp(const char* mesg); // add timestamp to timer
	void Reset();                 // reset timer
	void Print() const;           // print all timestamps
	float GetTotal() const;       // get total time (sec) since reset
	unsigned int GetMaxMem() const;// get maximum ram used (bytes) since reset

private:
	Timer();
	Timer(const Timer&);
	~Timer();
	long unsigned int CurrentTime() const;

	static Timer* m_pSingleton;

	std::vector<long unsigned int> m_time; // [0] is start time in milliseconds
	std::vector<std::string> m_mesg;  // description of timestamps
	unsigned int m_initialMemory; // initial memory, set at reset
};

#include <cstdio>
#include <sys/time.h>
#include "timer.h"
#if MEM_COUNTER
#if MAC_OS_X
#include <sys/malloc.h>
#else
#include <malloc.h>
#endif
#endif

Timer* Timer::m_pSingleton = NULL;

Timer* Timer::Get()
{
	if(!m_pSingleton) m_pSingleton = new Timer();
	return m_pSingleton;
}

void Timer::Destroy()
{
	if(m_pSingleton) delete m_pSingleton;
	m_pSingleton = NULL;
}

void Timer::Stamp(const char* mesg)
{
	m_time.push_back(CurrentTime());
	m_mesg.push_back(mesg);
//	printf("<TimeStamp> %s\n", mesg);
}

void Timer::Reset()
{
	m_time.clear();
	m_mesg.clear();

	m_time.push_back(CurrentTime());
	m_mesg.push_back("start");

#if MEM_COUNTER
	struct mallinfo info = mallinfo();
	m_initialMemory = info.uordblks; // current memory
#else
	m_initialMemory = 0;
#endif
}

void Timer::Print() const
{
	printf("Description                 | time  | elapsed\n");
	printf("----------------------------+-------+--------\n");

	int nbStamps = m_time.size();
	for(int i=1; i<nbStamps; i++)
	{
		float time = (m_time[i] - m_time[0]) / 1000.0;
		float elapsed = (m_time[i] - m_time[i-1]) / 1000.0;
		const char* mesg = m_mesg[i].c_str();

		printf("%-26s | %3.2f  | %3.2f\n", mesg, time, elapsed);
	}
	printf("----------------------------+-------+--------\n");
}

float Timer::GetTotal() const
{
	int nbStamps = m_time.size();
	float total = (m_time[nbStamps-1] - m_time[0]) / 1000.0;
	return total;
}

unsigned int Timer::GetMaxMem() const
{
#if MEM_COUNTER
	struct mallinfo info = mallinfo();
	int max = info.usmblks; /* replaced by LD_PRELOAD with max allocated memory*/
	return max - m_initialMemory;
#else
	return 0;
#endif
}

Timer::Timer()
{
	m_time.push_back(CurrentTime());
	m_mesg.push_back("start");

#if MEM_COUNTER
	struct mallinfo info = mallinfo();
	m_initialMemory = info.uordblks; // current memory
#else
	m_initialMemory = 0;
#endif
}

Timer::~Timer()
{
}

long unsigned int Timer::CurrentTime() const
{
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	long unsigned int t = tv.tv_sec*1000 + (long unsigned int)(tv.tv_usec/1000.0f);
	return t;
}

#endif
